#------------------------------------------------------------------------------

# MIT License

# Copyright (c) 2022 Landcare Research New Zealand Ltd

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

#------------------------------------------------------------------------------

# Python 3.7.4 (default, Aug  9 2019, 18:34:13) [MSC v.1915 64 bit (AMD64)]
import numpy as np # version 1.21.0
from nlmpy import nlmpy # version 1.0.1
import matplotlib as mpl # version 3.2.1

#------------------------------------------------------------------------------
# DEFINE BASIC PARAMETERS
#------------------------------------------------------------------------------

# Define domain for data
dim = 640
x = np.linspace(0, 1, dim)

# Define the parameters that control the waves
periods = 4
octaves = 3
lacunarity = 2
persistence = 0.5

#------------------------------------------------------------------------------
# GENERATE SINE WAVES
#------------------------------------------------------------------------------

# Function to calculate a sine wave
def sineWave(x, periods):
    y = np.sin(x * (2 * np.pi * periods))
    return(y)

# Generate a list of sine waves and a combined noise wave
for octave in range(octaves):
    if octave == 0:
        y = sineWave(x, periods = periods)
        sineOctaves = [y]
        noiseWave = np.copy(y)
    else:
        octavePeriod = periods * (octave * lacunarity)
        y = sineWave(x, periods = octavePeriod)
        y = y * (persistence ** octave)
        sineOctaves.append(y)
        noiseWave = noiseWave + y
sineOctaves.append(noiseWave)

#------------------------------------------------------------------------------
# GENERATE FOURIER SINE WAVES
#------------------------------------------------------------------------------

# Set random seed to get the same Fourier set
np.random.seed(7)

# Generate a list of sine waves and a combined noise wave
for octave in range(octaves):
    if octave == 0:
        y = sineWave(x, periods = 1)
        fourierOctaves = [y]
        fourierWave = np.copy(y)
    else:
        octavePeriod = 1 * (octave * lacunarity)
        y = sineWave(x + np.random.uniform(), periods = octavePeriod)
        y = y * (persistence ** octave)
        fourierOctaves.append(y)
        fourierWave = fourierWave + y
fourierOctaves.append(fourierWave)

#------------------------------------------------------------------------------
# GENERATE PERLIN NOISE
#------------------------------------------------------------------------------

# Set random seed to get the same Perlin noise
np.random.seed(2)

# Generate a list of Perlin noise surfaces and a combined Perlin noise surface
for octave in range(octaves):
    perlinOctave = nlmpy.octave(nRow = dim, nCol = dim, 
                                periods = (periods, periods),
                                persistence = persistence, 
                                octive = octave,
                                lacunarity = lacunarity)    
    if octave == 0:
        perlinOctaves = [perlinOctave]
        perlinNoise = np.copy(perlinOctave)
    else:
        perlinOctaves.append(perlinOctave)
        perlinNoise = perlinNoise + perlinOctave
perlinOctaves.append(perlinNoise)        

#------------------------------------------------------------------------------
# PLOTTING!
#------------------------------------------------------------------------------

mpl.pyplot.rcParams.update({'font.size': 7})
fig, axes = mpl.pyplot.subplots(nrows=4, ncols=4, figsize=(8, 6)) # figsize=(width, height) in inches
panels = ["(a)", "(b)", "(c)", "(d)", 
          "(e)", "(f)", "(g)", "(h)", 
          "(i)", "(j)", "(k)", "(l)",
          "(m)", "(n)", "(o)", "(p)"]
panelcount = -1

# Plot sine waves
ylim = 1.4
for octave in range(octaves + 1):
    axsi = axes[octave][0]
    axsi.plot(x, sineOctaves[octave], lw=0.75)
    axsi.set_ylim([-ylim, ylim])
    axsi.set_xlim([0, 1])
    axsi.yaxis.set_ticks([-ylim, -(ylim/2), 0, ylim/2, ylim])
    axsi.set_ylabel("Amplitude")
    axsi.axhline(0, lw=1, c='grey', ls=":")
    for i in range(1, periods):
        axsi.axvline(i / periods, lw=1, c='grey', ls=":")
    panelcount = panelcount + 1
    axsi.set_title(panels[panelcount], fontsize=7)
    if panelcount != 3:
        axsi.xaxis.set_ticks(np.linspace(0, 1, periods + 1))
        axsi.xaxis.set_ticklabels([])
    else:
        axsi.xaxis.set_ticks(np.linspace(0, 1, periods + 1))
        axsi.xaxis.set_ticklabels([0,1,2,3,4])
        axsi.set_xlabel("First sine wave periods")
    # Plot grey arrow
    if octave in [0,1,2]:
        aX = [0.2698, 0.7301, 0.7301, 0.2698]
        aY = [ylim, ylim, ylim + 0.5, ylim + 0.5]
        axsi.fill(aX, aY, ec=None, fc='silver', clip_on=False)
        aX = [0.2698, 0.7301, 0.7301, 0.2698]
        aY = [-ylim, -ylim, -ylim - 0.4, -ylim - 0.4]
        axsi.fill(aX, aY, ec=None, fc='silver', clip_on=False)           
    else:
        aX = [0.1697, 0.5, 0.8302]
        aY = [ylim + 0.5, ylim, ylim + 0.5]
        axsi.fill(aX, aY, ec=None, fc='silver', clip_on=False)



# Plot fourier waves
ylim = 1.4
for octave in range(octaves + 1):
    axsi = axes[octave][1]
    axsi.plot(x, fourierOctaves[octave], lw=0.75)
    axsi.set_ylim([-ylim, ylim])
    axsi.set_xlim([0, 1])
    axsi.yaxis.set_ticks([-ylim, -(ylim/2), 0, ylim/2, ylim])
    axsi.set_ylabel("")
    axsi.axhline(0, lw=1, c='grey', ls=":")
    for i in range(1, periods):
        axsi.axvline(i / periods, lw=1, c='grey', ls=":")
    panelcount = panelcount + 1
    axsi.set_title(panels[panelcount], fontsize=7)
    if panelcount != 7:
        axsi.xaxis.set_ticks(np.linspace(0, 1, periods + 1))
        axsi.xaxis.set_ticklabels([])
    else:
        axsi.xaxis.set_ticks(np.linspace(0, 1, periods + 1))
        axsi.xaxis.set_ticklabels(np.linspace(0, 1, periods + 1))
        axsi.set_xlabel("First sine wave period")
    # Plot grey arrow
    if octave in [0,1,2]:
        aX = [0.2698, 0.7301, 0.7301, 0.2698]
        aY = [ylim, ylim, ylim + 0.5, ylim + 0.5]
        axsi.fill(aX, aY, ec=None, fc='silver', clip_on=False)
        aX = [0.2698, 0.7301, 0.7301, 0.2698]
        aY = [-ylim, -ylim, -ylim - 0.4, -ylim - 0.4]
        axsi.fill(aX, aY, ec=None, fc='silver', clip_on=False)           
    else:
        aX = [0.1697, 0.5, 0.8302]
        aY = [ylim + 0.5, ylim, ylim + 0.5]
        axsi.fill(aX, aY, ec=None, fc='silver', clip_on=False)



# Plot Perlin noise
ylim = 0.8
for octave in range(octaves + 1):
    # Plot a cross-section
    axsi = axes[octave][2]
    axsi.plot(x, perlinOctaves[octave][160, :], lw=0.75)
    axsi.set_ylim([-ylim, ylim])
    axsi.set_xlim([0, 1])
    axsi.xaxis.set_ticks([])
    axsi.yaxis.set_ticks([-ylim, -(ylim/2), 0, ylim/2, ylim])
    axsi.axhline(0, lw=1, c='grey', ls=":")
    for i in range(1, periods):
        axsi.axvline(i / periods, lw=1, c='grey', ls=":")
    panelcount = panelcount + 1
    axsi.set_title(panels[panelcount], fontsize=7)
    if panelcount != 11:
        axsi.xaxis.set_ticks(np.linspace(0, 1, periods + 1))
        axsi.xaxis.set_ticklabels([])
    else:
        axsi.xaxis.set_ticks(np.linspace(0, 1, periods + 1))
        axsi.xaxis.set_ticklabels([0,1,2,3,4])
        axsi.set_xlabel("First octave periods")
    # Plot grey arrow
    if octave in [0,1,2]:
        aX = [0.2698, 0.7301, 0.7301, 0.2698]
        aY = [ylim, ylim, ylim + 0.2857, ylim + 0.2857]
        axsi.fill(aX, aY, ec=None, fc='silver', clip_on=False)
        aX = [0.2698, 0.7301, 0.7301, 0.2698]
        aY = [-ylim, -ylim, -ylim - 0.2285, -ylim - 0.2285]
        axsi.fill(aX, aY, ec=None, fc='silver', clip_on=False)
    else:
        aX = [0.1697, 0.5, 0.8302]
        aY = [ylim + 0.2857, ylim, ylim + 0.2857]
        axsi.fill(aX, aY, ec=None, fc='silver', clip_on=False)    
    
    # Plot the whole surface
    axsi = axes[octave][3]
    axsi.set_ylim([0, 4])
    axsi.set_xlim([0, 4])
    im = axsi.imshow(perlinOctaves[octave], vmin=-ylim, vmax=ylim, cmap=mpl.cm.plasma, extent=[0,4,0,4])
    cbar = fig.colorbar(im, ax=axsi, orientation='vertical', ticks=[-ylim, -(ylim/2), 0, ylim/2, ylim])
    cbar.set_label("Amplitude")    
    axsi.set_ylabel("First octave periods")
    if panelcount != 11:
        axsi.xaxis.set_ticks(np.linspace(0, 4, periods + 1))
        axsi.xaxis.set_ticklabels([])
    else:
        axsi.xaxis.set_ticks(np.linspace(0, 4, periods + 1))
        axsi.xaxis.set_ticklabels([0,1,2,3,4])
        axsi.set_xlabel("First octave periods")
    axsi.set_title(panels[panelcount + 4], fontsize=7)
    
    for i in range(1, periods):
        axsi.axvline(i, lw=1, c='grey', ls=":")
        axsi.axhline(i, lw=1, c='grey', ls=":")
    # Plot grey arrow
    if octave in [0,1,2]:
        aX = [0.8, 3.2, 3.2, 0.8]
        aY = [4, 4, 4 + 0.7142, 4 + 0.7142]
        axsi.fill(aX, aY, ec=None, fc='silver', clip_on=False)
        aX = [0.8, 3.2, 3.2, 0.8]
        aY = [0, 0, 0 - 0.5714, 0 - 0.5714]
        axsi.fill(aX, aY, ec=None, fc='silver', clip_on=False)    
    else:
        aX = [0.4, 2.0, 3.6]
        aY = [4 + 0.7142, 4, 4 + 0.7142]
        axsi.fill(aX, aY, ec=None, fc='silver', clip_on=False)    

#------------------------------------------------------------------------------

# Export figure
mpl.pyplot.tight_layout(pad=0.25)
mpl.pyplot.savefig('figure-1.png', dpi=300)

#------------------------------------------------------------------------------
